# Copyright (C) 2007, 2009, 2011, 2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for breezy.pack."""

from io import BytesIO

from ... import tests
from .. import pack


class TestContainerSerialiser(tests.TestCase):
    """Tests for the ContainerSerialiser class."""

    def test_construct(self):
        """Test constructing a ContainerSerialiser."""
        pack.ContainerSerialiser()

    def test_begin(self):
        serialiser = pack.ContainerSerialiser()
        self.assertEqual(b'Bazaar pack format 1 (introduced in 0.18)\n',
                         serialiser.begin())

    def test_end(self):
        serialiser = pack.ContainerSerialiser()
        self.assertEqual(b'E', serialiser.end())

    def test_bytes_record_no_name(self):
        serialiser = pack.ContainerSerialiser()
        record = serialiser.bytes_record(b'bytes', [])
        self.assertEqual(b'B5\n\nbytes', record)

    def test_bytes_record_one_name_with_one_part(self):
        serialiser = pack.ContainerSerialiser()
        record = serialiser.bytes_record(b'bytes', [(b'name',)])
        self.assertEqual(b'B5\nname\n\nbytes', record)

    def test_bytes_record_one_name_with_two_parts(self):
        serialiser = pack.ContainerSerialiser()
        record = serialiser.bytes_record(b'bytes', [(b'part1', b'part2')])
        self.assertEqual(b'B5\npart1\x00part2\n\nbytes', record)

    def test_bytes_record_two_names(self):
        serialiser = pack.ContainerSerialiser()
        record = serialiser.bytes_record(b'bytes', [(b'name1',), (b'name2',)])
        self.assertEqual(b'B5\nname1\nname2\n\nbytes', record)

    def test_bytes_record_whitespace_in_name_part(self):
        serialiser = pack.ContainerSerialiser()
        self.assertRaises(
            pack.InvalidRecordError,
            serialiser.bytes_record, b'bytes', [(b'bad name',)])

    def test_bytes_record_header(self):
        serialiser = pack.ContainerSerialiser()
        record = serialiser.bytes_header(32, [(b'name1',), (b'name2',)])
        self.assertEqual(b'B32\nname1\nname2\n\n', record)


class TestContainerWriter(tests.TestCase):

    def setUp(self):
        super().setUp()
        self.output = BytesIO()
        self.writer = pack.ContainerWriter(self.output.write)

    def assertOutput(self, expected_output):
        """Assert that the output of self.writer ContainerWriter is equal to
        expected_output.
        """
        self.assertEqual(expected_output, self.output.getvalue())

    def test_construct(self):
        """Test constructing a ContainerWriter.

        This uses None as the output stream to show that the constructor
        doesn't try to use the output stream.
        """
        pack.ContainerWriter(None)

    def test_begin(self):
        """The begin() method writes the container format marker line."""
        self.writer.begin()
        self.assertOutput(b'Bazaar pack format 1 (introduced in 0.18)\n')

    def test_zero_records_written_after_begin(self):
        """After begin is written, 0 records have been written."""
        self.writer.begin()
        self.assertEqual(0, self.writer.records_written)

    def test_end(self):
        """The end() method writes an End Marker record."""
        self.writer.begin()
        self.writer.end()
        self.assertOutput(b'Bazaar pack format 1 (introduced in 0.18)\nE')

    def test_empty_end_does_not_add_a_record_to_records_written(self):
        """The end() method does not count towards the records written."""
        self.writer.begin()
        self.writer.end()
        self.assertEqual(0, self.writer.records_written)

    def test_non_empty_end_does_not_add_a_record_to_records_written(self):
        """The end() method does not count towards the records written."""
        self.writer.begin()
        self.writer.add_bytes_record([b'foo'], len(b'foo'), names=[])
        self.writer.end()
        self.assertEqual(1, self.writer.records_written)

    def test_add_bytes_record_no_name(self):
        """Add a bytes record with no name."""
        self.writer.begin()
        offset, length = self.writer.add_bytes_record([b'abc'], len(b'abc'), names=[])
        self.assertEqual((42, 7), (offset, length))
        self.assertOutput(
            b'Bazaar pack format 1 (introduced in 0.18)\nB3\n\nabc')

    def test_add_bytes_record_one_name(self):
        """Add a bytes record with one name."""
        self.writer.begin()

        offset, length = self.writer.add_bytes_record(
            [b'abc'], len(b'abc'), names=[(b'name1', )])
        self.assertEqual((42, 13), (offset, length))
        self.assertOutput(
            b'Bazaar pack format 1 (introduced in 0.18)\n'
            b'B3\nname1\n\nabc')

    def test_add_bytes_record_split_writes(self):
        """Write a large record which does multiple IOs"""

        writes = []
        real_write = self.writer.write_func

        def record_writes(data):
            writes.append(data)
            return real_write(data)

        self.writer.write_func = record_writes
        self.writer._JOIN_WRITES_THRESHOLD = 2

        self.writer.begin()
        offset, length = self.writer.add_bytes_record(
            [b'abcabc'], len(b'abcabc'), names=[(b'name1', )])
        self.assertEqual((42, 16), (offset, length))
        self.assertOutput(
            b'Bazaar pack format 1 (introduced in 0.18)\n'
            b'B6\nname1\n\nabcabc')

        self.assertEqual([
            b'Bazaar pack format 1 (introduced in 0.18)\n',
            b'B6\nname1\n\n',
            b'abcabc'],
            writes)

    def test_add_bytes_record_two_names(self):
        """Add a bytes record with two names."""
        self.writer.begin()
        offset, length = self.writer.add_bytes_record(
            [b'abc'], len(b'abc'), names=[(b'name1', ), (b'name2', )])
        self.assertEqual((42, 19), (offset, length))
        self.assertOutput(
            b'Bazaar pack format 1 (introduced in 0.18)\n'
            b'B3\nname1\nname2\n\nabc')

    def test_add_bytes_record_two_element_name(self):
        """Add a bytes record with a two-element name."""
        self.writer.begin()
        offset, length = self.writer.add_bytes_record(
            [b'abc'], len(b'abc'), names=[(b'name1', b'name2')])
        self.assertEqual((42, 19), (offset, length))
        self.assertOutput(
            b'Bazaar pack format 1 (introduced in 0.18)\n'
            b'B3\nname1\x00name2\n\nabc')

    def test_add_second_bytes_record_gets_higher_offset(self):
        self.writer.begin()
        self.writer.add_bytes_record([b'a', b'bc'], len(b'abc'), names=[])
        offset, length = self.writer.add_bytes_record([b'abc'], len(b'abc'), names=[])
        self.assertEqual((49, 7), (offset, length))
        self.assertOutput(
            b'Bazaar pack format 1 (introduced in 0.18)\n'
            b'B3\n\nabc'
            b'B3\n\nabc')

    def test_add_bytes_record_invalid_name(self):
        """Adding a Bytes record with a name with whitespace in it raises
        InvalidRecordError.
        """
        self.writer.begin()
        self.assertRaises(
            pack.InvalidRecordError,
            self.writer.add_bytes_record, [b'abc'], len(b'abc'), names=[(b'bad name', )])

    def test_add_bytes_records_add_to_records_written(self):
        """Adding a Bytes record increments the records_written counter."""
        self.writer.begin()
        self.writer.add_bytes_record([b'foo'], len(b'foo'), names=[])
        self.assertEqual(1, self.writer.records_written)
        self.writer.add_bytes_record([b'foo'], len(b'foo'), names=[])
        self.assertEqual(2, self.writer.records_written)


class TestContainerReader(tests.TestCase):
    """Tests for the ContainerReader.

    The ContainerReader reads format 1 containers, so these tests explicitly
    test how it reacts to format 1 data.  If a new version of the format is
    added, then separate tests for that format should be added.
    """

    def get_reader_for(self, data):
        stream = BytesIO(data)
        reader = pack.ContainerReader(stream)
        return reader

    def test_construct(self):
        """Test constructing a ContainerReader.

        This uses None as the output stream to show that the constructor
        doesn't try to use the input stream.
        """
        pack.ContainerReader(None)

    def test_empty_container(self):
        """Read an empty container."""
        reader = self.get_reader_for(
            b"Bazaar pack format 1 (introduced in 0.18)\nE")
        self.assertEqual([], list(reader.iter_records()))

    def test_unknown_format(self):
        """Unrecognised container formats raise UnknownContainerFormatError."""
        reader = self.get_reader_for(b"unknown format\n")
        self.assertRaises(
            pack.UnknownContainerFormatError, reader.iter_records)

    def test_unexpected_end_of_container(self):
        """Containers that don't end with an End Marker record should cause
        UnexpectedEndOfContainerError to be raised.
        """
        reader = self.get_reader_for(
            b"Bazaar pack format 1 (introduced in 0.18)\n")
        iterator = reader.iter_records()
        self.assertRaises(
            pack.UnexpectedEndOfContainerError, next, iterator)

    def test_unknown_record_type(self):
        """Unknown record types cause UnknownRecordTypeError to be raised."""
        reader = self.get_reader_for(
            b"Bazaar pack format 1 (introduced in 0.18)\nX")
        iterator = reader.iter_records()
        self.assertRaises(
            pack.UnknownRecordTypeError, next, iterator)

    def test_container_with_one_unnamed_record(self):
        """Read a container with one Bytes record.

        Parsing Bytes records is more thoroughly exercised by
        TestBytesRecordReader.  This test is here to ensure that
        ContainerReader's integration with BytesRecordReader is working.
        """
        reader = self.get_reader_for(
            b"Bazaar pack format 1 (introduced in 0.18)\nB5\n\naaaaaE")
        expected_records = [([], b'aaaaa')]
        self.assertEqual(
            expected_records,
            [(names, read_bytes(None))
             for (names, read_bytes) in reader.iter_records()])

    def test_validate_empty_container(self):
        """validate does not raise an error for a container with no records."""
        reader = self.get_reader_for(
            b"Bazaar pack format 1 (introduced in 0.18)\nE")
        # No exception raised
        reader.validate()

    def test_validate_non_empty_valid_container(self):
        """validate does not raise an error for a container with a valid record.
        """
        reader = self.get_reader_for(
            b"Bazaar pack format 1 (introduced in 0.18)\nB3\nname\n\nabcE")
        # No exception raised
        reader.validate()

    def test_validate_bad_format(self):
        """validate raises an error for unrecognised format strings.

        It may raise either UnexpectedEndOfContainerError or
        UnknownContainerFormatError, depending on exactly what the string is.
        """
        inputs = [
            b"", b"x", b"Bazaar pack format 1 (introduced in 0.18)", b"bad\n"]
        for input in inputs:
            reader = self.get_reader_for(input)
            self.assertRaises(
                (pack.UnexpectedEndOfContainerError,
                 pack.UnknownContainerFormatError),
                reader.validate)

    def test_validate_bad_record_marker(self):
        """validate raises UnknownRecordTypeError for unrecognised record
        types.
        """
        reader = self.get_reader_for(
            b"Bazaar pack format 1 (introduced in 0.18)\nX")
        self.assertRaises(pack.UnknownRecordTypeError, reader.validate)

    def test_validate_data_after_end_marker(self):
        """validate raises ContainerHasExcessDataError if there are any bytes
        after the end of the container.
        """
        reader = self.get_reader_for(
            b"Bazaar pack format 1 (introduced in 0.18)\nEcrud")
        self.assertRaises(
            pack.ContainerHasExcessDataError, reader.validate)

    def test_validate_no_end_marker(self):
        """validate raises UnexpectedEndOfContainerError if there's no end of
        container marker, even if the container up to this point has been
        valid.
        """
        reader = self.get_reader_for(
            b"Bazaar pack format 1 (introduced in 0.18)\n")
        self.assertRaises(
            pack.UnexpectedEndOfContainerError, reader.validate)

    def test_validate_duplicate_name(self):
        """validate raises DuplicateRecordNameError if the same name occurs
        multiple times in the container.
        """
        reader = self.get_reader_for(
            b"Bazaar pack format 1 (introduced in 0.18)\n"
            b"B0\nname\n\n"
            b"B0\nname\n\n"
            b"E")
        self.assertRaises(pack.DuplicateRecordNameError, reader.validate)

    def test_validate_undecodeable_name(self):
        """Names that aren't valid UTF-8 cause validate to fail."""
        reader = self.get_reader_for(
            b"Bazaar pack format 1 (introduced in 0.18)\nB0\n\xcc\n\nE")
        self.assertRaises(pack.InvalidRecordError, reader.validate)


class TestBytesRecordReader(tests.TestCase):
    """Tests for reading and validating Bytes records with
    BytesRecordReader.

    Like TestContainerReader, this explicitly tests the reading of format 1
    data.  If a new version of the format is added, then a separate set of
    tests for reading that format should be added.
    """

    def get_reader_for(self, data):
        stream = BytesIO(data)
        reader = pack.BytesRecordReader(stream)
        return reader

    def test_record_with_no_name(self):
        """Reading a Bytes record with no name returns an empty list of
        names.
        """
        reader = self.get_reader_for(b"5\n\naaaaa")
        names, get_bytes = reader.read()
        self.assertEqual([], names)
        self.assertEqual(b'aaaaa', get_bytes(None))

    def test_record_with_one_name(self):
        """Reading a Bytes record with one name returns a list of just that
        name.
        """
        reader = self.get_reader_for(b"5\nname1\n\naaaaa")
        names, get_bytes = reader.read()
        self.assertEqual([(b'name1', )], names)
        self.assertEqual(b'aaaaa', get_bytes(None))

    def test_record_with_two_names(self):
        """Reading a Bytes record with two names returns a list of both names.
        """
        reader = self.get_reader_for(b"5\nname1\nname2\n\naaaaa")
        names, get_bytes = reader.read()
        self.assertEqual([(b'name1', ), (b'name2', )], names)
        self.assertEqual(b'aaaaa', get_bytes(None))

    def test_record_with_two_part_names(self):
        """Reading a Bytes record with a two_part name reads both."""
        reader = self.get_reader_for(b"5\nname1\x00name2\n\naaaaa")
        names, get_bytes = reader.read()
        self.assertEqual([(b'name1', b'name2', )], names)
        self.assertEqual(b'aaaaa', get_bytes(None))

    def test_invalid_length(self):
        """If the length-prefix is not a number, parsing raises
        InvalidRecordError.
        """
        reader = self.get_reader_for(b"not a number\n")
        self.assertRaises(pack.InvalidRecordError, reader.read)

    def test_early_eof(self):
        """Tests for premature EOF occuring during parsing Bytes records with
        BytesRecordReader.

        A incomplete container might be interrupted at any point.  The
        BytesRecordReader needs to cope with the input stream running out no
        matter where it is in the parsing process.

        In all cases, UnexpectedEndOfContainerError should be raised.
        """
        complete_record = b"6\nname\n\nabcdef"
        for count in range(0, len(complete_record)):
            incomplete_record = complete_record[:count]
            reader = self.get_reader_for(incomplete_record)
            # We don't use assertRaises to make diagnosing failures easier
            # (assertRaises doesn't allow a custom failure message).
            try:
                names, read_bytes = reader.read()
                read_bytes(None)
            except pack.UnexpectedEndOfContainerError:
                pass
            else:
                self.fail(
                    "UnexpectedEndOfContainerError not raised when parsing %r"
                    % (incomplete_record,))

    def test_initial_eof(self):
        """EOF before any bytes read at all."""
        reader = self.get_reader_for(b"")
        self.assertRaises(pack.UnexpectedEndOfContainerError, reader.read)

    def test_eof_after_length(self):
        """EOF after reading the length and before reading name(s)."""
        reader = self.get_reader_for(b"123\n")
        self.assertRaises(pack.UnexpectedEndOfContainerError, reader.read)

    def test_eof_during_name(self):
        """EOF during reading a name."""
        reader = self.get_reader_for(b"123\nname")
        self.assertRaises(pack.UnexpectedEndOfContainerError, reader.read)

    def test_read_invalid_name_whitespace(self):
        """Names must have no whitespace."""
        # A name with a space.
        reader = self.get_reader_for(b"0\nbad name\n\n")
        self.assertRaises(pack.InvalidRecordError, reader.read)

        # A name with a tab.
        reader = self.get_reader_for(b"0\nbad\tname\n\n")
        self.assertRaises(pack.InvalidRecordError, reader.read)

        # A name with a vertical tab.
        reader = self.get_reader_for(b"0\nbad\vname\n\n")
        self.assertRaises(pack.InvalidRecordError, reader.read)

    def test_validate_whitespace_in_name(self):
        """Names must have no whitespace."""
        reader = self.get_reader_for(b"0\nbad name\n\n")
        self.assertRaises(pack.InvalidRecordError, reader.validate)

    def test_validate_interrupted_prelude(self):
        """EOF during reading a record's prelude causes validate to fail."""
        reader = self.get_reader_for(b"")
        self.assertRaises(
            pack.UnexpectedEndOfContainerError, reader.validate)

    def test_validate_interrupted_body(self):
        """EOF during reading a record's body causes validate to fail."""
        reader = self.get_reader_for(b"1\n\n")
        self.assertRaises(
            pack.UnexpectedEndOfContainerError, reader.validate)

    def test_validate_unparseable_length(self):
        """An unparseable record length causes validate to fail."""
        reader = self.get_reader_for(b"\n\n")
        self.assertRaises(
            pack.InvalidRecordError, reader.validate)

    def test_validate_undecodeable_name(self):
        """Names that aren't valid UTF-8 cause validate to fail."""
        reader = self.get_reader_for(b"0\n\xcc\n\n")
        self.assertRaises(pack.InvalidRecordError, reader.validate)

    def test_read_max_length(self):
        """If the max_length passed to the callable returned by read is not
        None, then no more than that many bytes will be read.
        """
        reader = self.get_reader_for(b"6\n\nabcdef")
        names, get_bytes = reader.read()
        self.assertEqual(b'abc', get_bytes(3))

    def test_read_no_max_length(self):
        """If the max_length passed to the callable returned by read is None,
        then all the bytes in the record will be read.
        """
        reader = self.get_reader_for(b"6\n\nabcdef")
        names, get_bytes = reader.read()
        self.assertEqual(b'abcdef', get_bytes(None))

    def test_repeated_read_calls(self):
        """Repeated calls to the callable returned from BytesRecordReader.read
        will not read beyond the end of the record.
        """
        reader = self.get_reader_for(b"6\n\nabcdefB3\nnext-record\nXXX")
        names, get_bytes = reader.read()
        self.assertEqual(b'abcdef', get_bytes(None))
        self.assertEqual(b'', get_bytes(None))
        self.assertEqual(b'', get_bytes(99))


class TestMakeReadvReader(tests.TestCaseWithTransport):

    def test_read_skipping_records(self):
        pack_data = BytesIO()
        writer = pack.ContainerWriter(pack_data.write)
        writer.begin()
        memos = []
        memos.append(writer.add_bytes_record([b'abc'], 3, names=[]))
        memos.append(writer.add_bytes_record([b'def'], 3, names=[(b'name1', )]))
        memos.append(writer.add_bytes_record([b'ghi'], 3, names=[(b'name2', )]))
        memos.append(writer.add_bytes_record([b'jkl'], 3, names=[]))
        writer.end()
        transport = self.get_transport()
        transport.put_bytes('mypack', pack_data.getvalue())
        requested_records = [memos[0], memos[2]]
        reader = pack.make_readv_reader(transport, 'mypack', requested_records)
        result = []
        for names, reader_func in reader.iter_records():
            result.append((names, reader_func(None)))
        self.assertEqual([([], b'abc'), ([(b'name2', )], b'ghi')], result)


class TestReadvFile(tests.TestCaseWithTransport):
    """Tests of the ReadVFile class.

    Error cases are deliberately undefined: this code adapts the underlying
    transport interface to a single 'streaming read' interface as
    ContainerReader needs.
    """

    def test_read_bytes(self):
        """Test reading of both single bytes and all bytes in a hunk."""
        transport = self.get_transport()
        transport.put_bytes('sample', b'0123456789')
        f = pack.ReadVFile(transport.readv(
            'sample', [(0, 1), (1, 2), (4, 1), (6, 2)]))
        results = []
        results.append(f.read(1))
        results.append(f.read(2))
        results.append(f.read(1))
        results.append(f.read(1))
        results.append(f.read(1))
        self.assertEqual([b'0', b'12', b'4', b'6', b'7'], results)

    def test_readline(self):
        """Test using readline() as ContainerReader does.

        This is always within a readv hunk, never across it.
        """
        transport = self.get_transport()
        transport.put_bytes('sample', b'0\n2\n4\n')
        f = pack.ReadVFile(transport.readv('sample', [(0, 2), (2, 4)]))
        results = []
        results.append(f.readline())
        results.append(f.readline())
        results.append(f.readline())
        self.assertEqual([b'0\n', b'2\n', b'4\n'], results)

    def test_readline_and_read(self):
        """Test exercising one byte reads, readline, and then read again."""
        transport = self.get_transport()
        transport.put_bytes('sample', b'0\n2\n4\n')
        f = pack.ReadVFile(transport.readv('sample', [(0, 6)]))
        results = []
        results.append(f.read(1))
        results.append(f.readline())
        results.append(f.read(4))
        self.assertEqual([b'0', b'\n', b'2\n4\n'], results)


class PushParserTestCase(tests.TestCase):
    """Base class for TestCases involving ContainerPushParser."""

    def make_parser_expecting_record_type(self):
        parser = pack.ContainerPushParser()
        parser.accept_bytes(b"Bazaar pack format 1 (introduced in 0.18)\n")
        return parser

    def make_parser_expecting_bytes_record(self):
        parser = pack.ContainerPushParser()
        parser.accept_bytes(b"Bazaar pack format 1 (introduced in 0.18)\nB")
        return parser

    def assertRecordParsing(self, expected_record, data):
        """Assert that 'bytes' is parsed as a given bytes record.

        :param expected_record: A tuple of (names, bytes).
        """
        parser = self.make_parser_expecting_bytes_record()
        parser.accept_bytes(data)
        parsed_records = parser.read_pending_records()
        self.assertEqual([expected_record], parsed_records)


class TestContainerPushParser(PushParserTestCase):
    """Tests for ContainerPushParser.

    The ContainerPushParser reads format 1 containers, so these tests
    explicitly test how it reacts to format 1 data.  If a new version of the
    format is added, then separate tests for that format should be added.
    """

    def test_construct(self):
        """ContainerPushParser can be constructed."""
        pack.ContainerPushParser()

    def test_multiple_records_at_once(self):
        """If multiple records worth of data are fed to the parser in one
        string, the parser will correctly parse all the records.

        (A naive implementation might stop after parsing the first record.)
        """
        parser = self.make_parser_expecting_record_type()
        parser.accept_bytes(b"B5\nname1\n\nbody1B5\nname2\n\nbody2")
        self.assertEqual(
            [([(b'name1',)], b'body1'), ([(b'name2',)], b'body2')],
            parser.read_pending_records())

    def test_multiple_empty_records_at_once(self):
        """If multiple empty records worth of data are fed to the parser in one
        string, the parser will correctly parse all the records.

        (A naive implementation might stop after parsing the first empty
        record, because the buffer size had not changed.)
        """
        parser = self.make_parser_expecting_record_type()
        parser.accept_bytes(b"B0\nname1\n\nB0\nname2\n\n")
        self.assertEqual(
            [([(b'name1',)], b''), ([(b'name2',)], b'')],
            parser.read_pending_records())


class TestContainerPushParserBytesParsing(PushParserTestCase):
    """Tests for reading Bytes records with ContainerPushParser.

    The ContainerPushParser reads format 1 containers, so these tests
    explicitly test how it reacts to format 1 data.  If a new version of the
    format is added, then separate tests for that format should be added.
    """

    def test_record_with_no_name(self):
        """Reading a Bytes record with no name returns an empty list of
        names.
        """
        self.assertRecordParsing(([], b'aaaaa'), b"5\n\naaaaa")

    def test_record_with_one_name(self):
        """Reading a Bytes record with one name returns a list of just that
        name.
        """
        self.assertRecordParsing(
            ([(b'name1', )], b'aaaaa'),
            b"5\nname1\n\naaaaa")

    def test_record_with_two_names(self):
        """Reading a Bytes record with two names returns a list of both names.
        """
        self.assertRecordParsing(
            ([(b'name1', ), (b'name2', )], b'aaaaa'),
            b"5\nname1\nname2\n\naaaaa")

    def test_record_with_two_part_names(self):
        """Reading a Bytes record with a two_part name reads both."""
        self.assertRecordParsing(
            ([(b'name1', b'name2')], b'aaaaa'),
            b"5\nname1\x00name2\n\naaaaa")

    def test_invalid_length(self):
        """If the length-prefix is not a number, parsing raises
        InvalidRecordError.
        """
        parser = self.make_parser_expecting_bytes_record()
        self.assertRaises(
            pack.InvalidRecordError, parser.accept_bytes, b"not a number\n")

    def test_incomplete_record(self):
        """If the bytes seen so far don't form a complete record, then there
        will be nothing returned by read_pending_records.
        """
        parser = self.make_parser_expecting_bytes_record()
        parser.accept_bytes(b"5\n\nabcd")
        self.assertEqual([], parser.read_pending_records())

    def test_accept_nothing(self):
        """The edge case of parsing an empty string causes no error."""
        parser = self.make_parser_expecting_bytes_record()
        parser.accept_bytes(b"")

    def assertInvalidRecord(self, data):
        """Assert that parsing the given bytes raises InvalidRecordError."""
        parser = self.make_parser_expecting_bytes_record()
        self.assertRaises(
            pack.InvalidRecordError, parser.accept_bytes, data)

    def test_read_invalid_name_whitespace(self):
        """Names must have no whitespace."""
        # A name with a space.
        self.assertInvalidRecord(b"0\nbad name\n\n")

        # A name with a tab.
        self.assertInvalidRecord(b"0\nbad\tname\n\n")

        # A name with a vertical tab.
        self.assertInvalidRecord(b"0\nbad\vname\n\n")

    def test_repeated_read_pending_records(self):
        """read_pending_records will not return the same record twice."""
        parser = self.make_parser_expecting_bytes_record()
        parser.accept_bytes(b"6\n\nabcdef")
        self.assertEqual([([], b'abcdef')], parser.read_pending_records())
        self.assertEqual([], parser.read_pending_records())


class TestErrors(tests.TestCase):

    def test_unknown_container_format(self):
        """Test the formatting of UnknownContainerFormatError."""
        e = pack.UnknownContainerFormatError('bad format string')
        self.assertEqual(
            "Unrecognised container format: 'bad format string'",
            str(e))

    def test_unexpected_end_of_container(self):
        """Test the formatting of UnexpectedEndOfContainerError."""
        e = pack.UnexpectedEndOfContainerError()
        self.assertEqual(
            "Unexpected end of container stream", str(e))

    def test_unknown_record_type(self):
        """Test the formatting of UnknownRecordTypeError."""
        e = pack.UnknownRecordTypeError("X")
        self.assertEqual(
            "Unknown record type: 'X'",
            str(e))

    def test_invalid_record(self):
        """Test the formatting of InvalidRecordError."""
        e = pack.InvalidRecordError("xxx")
        self.assertEqual(
            "Invalid record: xxx",
            str(e))

    def test_container_has_excess_data(self):
        """Test the formatting of ContainerHasExcessDataError."""
        e = pack.ContainerHasExcessDataError("excess bytes")
        self.assertEqual(
            "Container has data after end marker: 'excess bytes'",
            str(e))

    def test_duplicate_record_name_error(self):
        """Test the formatting of DuplicateRecordNameError."""
        e = pack.DuplicateRecordNameError(b"n\xc3\xa5me")
        self.assertEqual(
            "Container has multiple records with the same name: n\xe5me",
            str(e))



# Copyright (C) 2007-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for fetch between repositories of the same type."""

from breezy import controldir, errors, gpg, repository
from breezy.bzr import remote
from breezy.bzr.inventory import ROOT_ID
from breezy.tests import TestNotApplicable, TestSkipped
from breezy.tests.per_repository import TestCaseWithRepository


class TestFetchSameRepository(TestCaseWithRepository):

    def test_fetch(self):
        # smoke test fetch to ensure that the convenience function works.
        # it is defined as a convenience function with the underlying
        # functionality provided by an InterRepository
        tree_a = self.make_branch_and_tree('a')
        self.build_tree(['a/foo'])
        tree_a.add('foo')
        rev1 = tree_a.commit('rev1')
        # fetch with a default limit (grab everything)
        repo = self.make_repository('b')
        if (tree_a.branch.repository.supports_rich_root() and not
                repo.supports_rich_root()):
            raise TestSkipped('Cannot fetch from model2 to model1')
        repo.fetch(tree_a.branch.repository,
                   revision_id=None)

    def test_fetch_fails_in_write_group(self):
        # fetch() manages a write group itself, fetching within one isn't safe.
        repo = self.make_repository('a')
        repo.lock_write()
        self.addCleanup(repo.unlock)
        repo.start_write_group()
        self.addCleanup(repo.abort_write_group)
        # Don't need a specific class - not expecting flow control based on
        # this.
        self.assertRaises(errors.BzrError, repo.fetch, repo)

    def test_fetch_to_knit3(self):
        # create a repository of the sort we are testing.
        tree_a = self.make_branch_and_tree('a')
        self.build_tree(['a/foo'])
        tree_a.add('foo')
        rev1 = tree_a.commit('rev1')
        # create a knit-3 based format to fetch into
        f = controldir.format_registry.make_controldir('development-subtree')
        try:
            format = tree_a.branch.repository._format
            format.check_conversion_target(f.repository_format)
            # if we cannot convert data to knit3, skip the test.
        except errors.BadConversionTarget as e:
            raise TestSkipped(str(e))
        self.get_transport().mkdir('b')
        b_bzrdir = f.initialize(self.get_url('b'))
        knit3_repo = b_bzrdir.create_repository()
        # fetch with a default limit (grab everything)
        knit3_repo.fetch(tree_a.branch.repository, revision_id=None)
        # Reopen to avoid any in-memory caching - ensure its reading from
        # disk.
        knit3_repo = b_bzrdir.open_repository()
        rev1_tree = knit3_repo.revision_tree(rev1)
        with rev1_tree.lock_read():
            lines = rev1_tree.get_file_lines('')
        self.assertEqual([], lines)
        b_branch = b_bzrdir.create_branch()
        b_branch.pull(tree_a.branch)
        try:
            tree_b = b_bzrdir.create_workingtree()
        except errors.NotLocalUrl:
            try:
                tree_b = b_branch.create_checkout('b', lightweight=True)
            except errors.NotLocalUrl:
                raise TestSkipped("cannot make working tree with transport %r"
                                  % b_bzrdir.transport)
        rev2 = tree_b.commit('no change')
        rev2_tree = knit3_repo.revision_tree(rev2)
        self.assertEqual(rev1, rev2_tree.get_file_revision(''))

    def do_test_fetch_to_rich_root_sets_parents_correctly(self, result,
                                                          snapshots, root_id=ROOT_ID, allow_lefthand_ghost=False):
        """Assert that result is the parents of b'tip' after fetching snapshots.

        This helper constructs a 1.9 format source, and a test-format target
        and fetches the result of building snapshots in the source, then
        asserts that the parents of tip are result.

        :param result: A parents list for the inventories.get_parent_map call.
        :param snapshots: An iterable of snapshot parameters for
            BranchBuilder.build_snapshot.
        '"""
        # This overlaps slightly with the tests for commit builder about graph
        # consistency.
        # Cases:
        repo = self.make_repository('target')
        remote_format = isinstance(repo, remote.RemoteRepository)
        if not repo._format.rich_root_data and not remote_format:
            return  # not relevant
        if not repo._format.supports_full_versioned_files:
            raise TestNotApplicable(
                'format does not support full versioned files')
        builder = self.make_branch_builder('source', format='1.9')
        builder.start_series()
        for revision_id, parent_ids, actions in snapshots:
            builder.build_snapshot(parent_ids, actions,
                                   allow_leftmost_as_ghost=allow_lefthand_ghost,
                                   revision_id=revision_id)
        builder.finish_series()
        source = builder.get_branch()
        if remote_format and not repo._format.rich_root_data:
            # use a manual rich root format to ensure the code path is tested.
            repo = self.make_repository('remote-target',
                                        format='1.9-rich-root')
        repo.lock_write()
        self.addCleanup(repo.unlock)
        repo.fetch(source.repository)
        graph = repo.get_file_graph()
        self.assertEqual(result,
                         graph.get_parent_map([(root_id, b'tip')])[(root_id, b'tip')])

    def test_fetch_to_rich_root_set_parent_no_parents(self):
        # No parents rev -> No parents
        self.do_test_fetch_to_rich_root_sets_parents_correctly((),
                                                               [(b'tip', None, [('add', ('', ROOT_ID, 'directory', ''))]),
                                                                ])

    def test_fetch_to_rich_root_set_parent_1_parent(self):
        # 1 parent rev -> 1 parent
        self.do_test_fetch_to_rich_root_sets_parents_correctly(
            ((ROOT_ID, b'base'),),
            [(b'base', None, [('add', ('', ROOT_ID, 'directory', ''))]),
             (b'tip', None, []),
             ])

    def test_fetch_to_rich_root_set_parent_1_ghost_parent(self):
        # 1 ghost parent -> No parents
        if not self.repository_format.supports_ghosts:
            raise TestNotApplicable("repository format does not support "
                                    "ghosts")
        self.do_test_fetch_to_rich_root_sets_parents_correctly((),
                                                               [(b'tip', [b'ghost'], [('add', ('', ROOT_ID, 'directory', ''))]),
                                                                ], allow_lefthand_ghost=True)

    def test_fetch_to_rich_root_set_parent_2_head_parents(self):
        # 2 parents both heads -> 2 parents
        self.do_test_fetch_to_rich_root_sets_parents_correctly(
            ((ROOT_ID, b'left'), (ROOT_ID, b'right')),
            [(b'base', None, [('add', ('', ROOT_ID, 'directory', ''))]),
             (b'left', None, []),
             (b'right', [b'base'], []),
             (b'tip', [b'left', b'right'], []),
             ])

    def test_fetch_to_rich_root_set_parent_2_parents_1_head(self):
        # 2 parents one head -> 1 parent
        self.do_test_fetch_to_rich_root_sets_parents_correctly(
            ((ROOT_ID, b'right'),),
            [(b'left', None, [('add', ('', ROOT_ID, 'directory', ''))]),
             (b'right', None, []),
             (b'tip', [b'left', b'right'], []),
             ])

    def test_fetch_to_rich_root_set_parent_1_parent_different_id_gone(self):
        # 1 parent different fileid, ours missing -> no parents
        self.do_test_fetch_to_rich_root_sets_parents_correctly(
            (),
            [(b'base', None, [('add', ('', ROOT_ID, 'directory', ''))]),
             (b'tip', None, [('unversion', ''),
                             ('add', ('', b'my-root', 'directory', '')),
                             ]),
             ], root_id=b'my-root')

    def test_fetch_to_rich_root_set_parent_1_parent_different_id_moved(self):
        # 1 parent different fileid, ours moved -> 1 parent
        # (and that parent honours the changing revid of the other location)
        self.do_test_fetch_to_rich_root_sets_parents_correctly(
            ((b'my-root', b'origin'),),
            [(b'origin', None, [('add', ('', ROOT_ID, 'directory', '')),
                                ('add', ('child', b'my-root', 'directory', ''))]),
             (b'base', None, []),
             (b'tip', None, [('unversion', 'child'),
                             ('unversion', ''),
                             ('flush', None),
                             ('add', ('', b'my-root', 'directory', '')),
                             ]),
             ], root_id=b'my-root')

    def test_fetch_to_rich_root_set_parent_2_parent_1_different_id_gone(self):
        # 2 parents, 1 different fileid, our second missing -> 1 parent
        self.do_test_fetch_to_rich_root_sets_parents_correctly(
            ((b'my-root', b'right'),),
            [(b'base', None, [('add', ('', ROOT_ID, 'directory', ''))]),
             (b'right', None, [('unversion', ''),
                               ('add', ('', b'my-root', 'directory', ''))]),
             (b'tip', [b'base', b'right'], [('unversion', ''),
                                            ('add', ('', b'my-root', 'directory', '')),
                                            ]),
             ], root_id=b'my-root')

    def test_fetch_to_rich_root_set_parent_2_parent_2_different_id_moved(self):
        # 2 parents, 1 different fileid, our second moved -> 2 parent
        # (and that parent honours the changing revid of the other location)
        self.do_test_fetch_to_rich_root_sets_parents_correctly(
            ((b'my-root', b'right'),),
            # b'my-root' at 'child'.
            [(b'origin', None, [('add', ('', ROOT_ID, 'directory', '')),
                                ('add', ('child', b'my-root', 'directory', ''))]),
             (b'base', None, []),
             # b'my-root' at root
             (b'right', None, [('unversion', 'child'),
                               ('unversion', ''),
                               ('flush', None),
                               ('add', ('', b'my-root', 'directory', ''))]),
             (b'tip', [b'base', b'right'], [('unversion', ''),
                                            ('unversion', 'child'),
                                            ('flush', None),
                                            ('add', ('', b'my-root', 'directory', '')),
                                            ]),
             ], root_id=b'my-root')

    def test_fetch_all_from_self(self):
        tree = self.make_branch_and_tree('.')
        rev_id = tree.commit('one')
        # This needs to be a new copy of the repository, if this changes, the
        # test needs to be rewritten
        repo = tree.branch.repository.controldir.open_repository()
        # This fetch should be a no-op see bug #158333
        tree.branch.repository.fetch(repo, None)

    def test_fetch_from_self(self):
        tree = self.make_branch_and_tree('.')
        rev_id = tree.commit('one')
        repo = tree.branch.repository.controldir.open_repository()
        # This fetch should be a no-op see bug #158333
        tree.branch.repository.fetch(repo, rev_id)

    def test_fetch_missing_from_self(self):
        tree = self.make_branch_and_tree('.')
        rev_id = tree.commit('one')
        # Even though the fetch() is a NO-OP it should assert the revision id
        # is present
        repo = tree.branch.repository.controldir.open_repository()
        self.assertRaises(errors.NoSuchRevision, tree.branch.repository.fetch,
                          repo, b'no-such-revision')

    def makeARepoWithSignatures(self):
        wt = self.make_branch_and_tree('a-repo-with-sigs')
        rev1 = wt.commit('rev1', allow_pointless=True)
        repo = wt.branch.repository
        repo.lock_write()
        repo.start_write_group()
        try:
            repo.sign_revision(rev1, gpg.LoopbackGPGStrategy(None))
        except errors.UnsupportedOperation:
            self.assertFalse(repo._format.supports_revision_signatures)
            raise TestNotApplicable(
                "repository format does not support signatures")
        repo.commit_write_group()
        repo.unlock()
        return repo, rev1

    def test_fetch_copies_signatures(self):
        source_repo, rev1 = self.makeARepoWithSignatures()
        target_repo = self.make_repository('target')
        target_repo.fetch(source_repo, revision_id=None)
        self.assertEqual(
            source_repo.get_signature_text(rev1),
            target_repo.get_signature_text(rev1))

    def make_repository_with_one_revision(self):
        wt = self.make_branch_and_tree('source')
        rev1 = wt.commit('rev1', allow_pointless=True)
        return wt.branch.repository, rev1

    def test_fetch_revision_already_exists(self):
        # Make a repository with one revision.
        source_repo, rev1 = self.make_repository_with_one_revision()
        # Fetch that revision into a second repository.
        target_repo = self.make_repository('target')
        target_repo.fetch(source_repo, revision_id=rev1)
        # Now fetch again; there will be nothing to do.  This should work
        # without causing any errors.
        target_repo.fetch(source_repo, revision_id=rev1)

    def test_fetch_all_same_revisions_twice(self):
        # Blind-fetching all the same revisions twice should succeed and be a
        # no-op the second time.
        repo = self.make_repository('repo')
        tree = self.make_branch_and_tree('tree')
        revision_id = tree.commit('test')
        repo.fetch(tree.branch.repository)
        repo.fetch(tree.branch.repository)

    def make_simple_branch_with_ghost(self):
        if not self.repository_format.supports_ghosts:
            raise TestNotApplicable("repository format does not support "
                                    "ghosts")
        builder = self.make_branch_builder('source')
        builder.start_series()
        a_revid = builder.build_snapshot(None, [
            ('add', ('', b'root-id', 'directory', None)),
            ('add', ('file', b'file-id', 'file', b'content\n'))])
        b_revid = builder.build_snapshot([a_revid, b'ghost-id'], [])
        builder.finish_series()
        source_b = builder.get_branch()
        source_b.lock_read()
        self.addCleanup(source_b.unlock)
        return source_b, b_revid

    def test_fetch_with_ghost(self):
        source_b, b_revid = self.make_simple_branch_with_ghost()
        target = self.make_repository('target')
        target.lock_write()
        self.addCleanup(target.unlock)
        target.fetch(source_b.repository, revision_id=b_revid)

    def test_fetch_into_smart_with_ghost(self):
        trans = self.make_smart_server('target')
        source_b, b_revid = self.make_simple_branch_with_ghost()
        if not source_b.controldir._format.supports_transport(trans):
            raise TestNotApplicable("format does not support transport")
        target = self.make_repository('target')
        # Re-open the repository over the smart protocol
        target = repository.Repository.open(trans.base)
        target.lock_write()
        self.addCleanup(target.unlock)
        try:
            target.fetch(source_b.repository, revision_id=b_revid)
        except errors.TokenLockingNotSupported:
            # The code inside fetch() that tries to lock and then fails, also
            # causes weird problems with 'lock_not_held' later on...
            target.lock_read()
            self.knownFailure('some repositories fail to fetch'
                              ' via the smart server because of locking issues.')

    def test_fetch_from_smart_with_ghost(self):
        trans = self.make_smart_server('source')
        source_b, b_revid = self.make_simple_branch_with_ghost()
        if not source_b.controldir._format.supports_transport(trans):
            raise TestNotApplicable("format does not support transport")
        target = self.make_repository('target')
        target.lock_write()
        self.addCleanup(target.unlock)
        # Re-open the repository over the smart protocol
        source = repository.Repository.open(trans.base)
        source.lock_read()
        self.addCleanup(source.unlock)
        target.fetch(source, revision_id=b_revid)



# Copyright (C) 2006-2012 Aaron Bentley
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os
import tarfile
import tempfile
import warnings
from io import BytesIO
from shutil import copy2, copytree, rmtree

from .. import osutils
from .. import revision as _mod_revision
from .. import transform
from ..controldir import ControlDir
from ..export import export
from ..upstream_import import (NotArchiveType, ZipFileWrapper,
                               common_directory, get_archive_type,
                               import_archive, import_dir, import_tar,
                               import_zip, top_path)
from . import TestCaseInTempDir, TestCaseWithTransport
from .features import UnicodeFilenameFeature


def import_tar_broken(tree, tar_input):
    """
    Import a tarfile with names that that end in //, e.g. Feisty Python 2.5
    """
    tar_file = tarfile.open('lala', 'r', tar_input)
    for member in tar_file.members:
        if member.name.endswith('/'):
            member.name += '/'
    import_archive(tree, tar_file)


class DirFileWriter:

    def __init__(self, fileobj, mode):
        # We may be asked to 'append'.  If so, fileobj already has a path.
        # So we copy the existing tree, and overwrite afterward.
        fileobj.seek(0)
        existing = fileobj.read()
        fileobj.seek(0)
        path = tempfile.mkdtemp(dir=os.getcwd())
        if existing != b'':
            # copytree requires the directory not to exist
            os.rmdir(path)
            copytree(existing, path)
        fileobj.write(path.encode('utf-8'))
        self.root = path

    def add(self, path):
        target_path = os.path.join(self.root, path)
        parent = osutils.dirname(target_path)
        if not os.path.exists(parent):
            os.makedirs(parent)
        kind = osutils.file_kind(path)
        if kind == 'file':
            copy2(path, target_path)
        if kind == 'directory':
            os.mkdir(target_path)

    def close(self):
        pass


class TestImport(TestCaseInTempDir):

    def make_tar(self, mode='w'):
        def maker(fileobj):
            return tarfile.open('project-0.1.tar', mode, fileobj)
        return self.make_archive(maker)

    def make_archive(self, maker, subdir=True):
        result = BytesIO()
        archive_file = maker(result)
        try:
            os.mkdir('project-0.1')
            if subdir:
                prefix = 'project-0.1/'
                archive_file.add('project-0.1')
            else:
                prefix = ''
                os.chdir('project-0.1')
            os.mkdir(prefix + 'junk')
            archive_file.add(prefix + 'junk')

            with open(prefix + 'README', 'wb') as f:
                f.write(b'What?')
            archive_file.add(prefix + 'README')

            with open(prefix + 'FEEDME', 'wb') as f:
                f.write(b'Hungry!!')
            archive_file.add(prefix + 'FEEDME')

            archive_file.close()
        finally:
            if not subdir:
                os.chdir('..')
            rmtree('project-0.1')
        result.seek(0)
        return result

    def make_archive2(self, builder, subdir):
        result = BytesIO()
        archive_file = builder(result)
        os.mkdir('project-0.2')
        try:
            if subdir:
                prefix = 'project-0.2/'
                archive_file.add('project-0.2')
            else:
                prefix = ''
                os.chdir('project-0.2')

            os.mkdir(prefix + 'junk')
            archive_file.add(prefix + 'junk')

            with open(prefix + 'README', 'wb') as f:
                f.write(b'Now?')
            archive_file.add(prefix + 'README')

            with open(prefix + 'README', 'wb') as f:
                f.write(b'Wow?')
            # Add a second entry for README with different contents.
            archive_file.add(prefix + 'README')
            archive_file.close()

        finally:
            if not subdir:
                os.chdir('..')
        result.seek(0)
        return result

    def make_messed_tar(self):
        result = BytesIO()
        with tarfile.open('project-0.1.tar', 'w', result) as tar_file:
            os.mkdir('project-0.1')
            tar_file.add('project-0.1')

            os.mkdir('project-0.2')
            tar_file.add('project-0.2')

            with open('project-0.1/README', 'wb') as f:
                f.write(b'What?')
            tar_file.add('project-0.1/README')
        rmtree('project-0.1')
        result.seek(0)
        return result

    def make_zip(self):
        def maker(fileobj):
            return ZipFileWrapper(fileobj, 'w')
        return self.make_archive(maker)

    def make_tar_with_bzrdir(self):
        result = BytesIO()
        with tarfile.open('tar-with-bzrdir.tar', 'w', result) as tar_file:
            os.mkdir('toplevel-dir')
            tar_file.add('toplevel-dir')
            os.mkdir('toplevel-dir/.bzr')
            tar_file.add('toplevel-dir/.bzr')
        rmtree('toplevel-dir')
        result.seek(0)
        return result

    def test_top_path(self):
        self.assertEqual(top_path('ab/b/c'), 'ab')
        self.assertEqual(top_path('etc'), 'etc')
        self.assertEqual(top_path('project-0.1'), 'project-0.1')

    def test_common_directory(self):
        self.assertEqual(common_directory(['ab/c/d', 'ab/c/e']), 'ab')
        self.assertIs(common_directory(['ab/c/d', 'ac/c/e']), None)
        self.assertEqual('FEEDME', common_directory(['FEEDME']))

    def test_untar(self):
        def builder(fileobj, mode='w'):
            return tarfile.open('project-0.1.tar', mode, fileobj)
        self.archive_test(builder, import_tar)

    def test_broken_tar(self):
        def builder(fileobj, mode='w'):
            return tarfile.open('project-0.1.tar', mode, fileobj)
        self.archive_test(builder, import_tar_broken, subdir=True)

    def test_unzip(self):
        def builder(fileobj, mode='w'):
            return ZipFileWrapper(fileobj, mode)
        self.archive_test(builder, import_zip)

    def test_copydir_nosub(self):
        def builder(fileobj, mode='w'):
            return DirFileWriter(fileobj, mode)
        # It would be bogus to test with the result in a subdirectory,
        # because for directories, the input root is always the output root.
        self.archive_test(builder, import_dir)

    def archive_test(self, builder, importer, subdir=False):
        archive_file = self.make_archive(builder, subdir)
        tree = ControlDir.create_standalone_workingtree('tree')
        with tree.lock_write():
            importer(tree, archive_file)
            self.assertTrue(tree.is_versioned('README'))
            self.assertTrue(tree.is_versioned('FEEDME'))
            self.assertTrue(os.path.isfile(tree.abspath('README')))
            self.assertEqual(tree.stored_kind('README'), 'file')
            self.assertEqual(tree.stored_kind('FEEDME'), 'file')
            with open(tree.abspath('junk/food'), 'wb') as f:
                f.write(b'I like food\n')

            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                archive_file = self.make_archive2(builder, subdir)
                importer(tree, archive_file)
            self.assertTrue(tree.is_versioned('README'))
            # Ensure the second version of the file is used.
            self.assertEqual(tree.get_file_text('README'), b'Wow?')
            self.assertTrue(not os.path.exists(tree.abspath('FEEDME')))

    def test_untar2(self):
        tar_file = self.make_messed_tar()
        tree = ControlDir.create_standalone_workingtree('tree')
        import_tar(tree, tar_file)
        self.assertTrue(tree.is_versioned('project-0.1/README'))

    def test_untar_gzip(self):
        tar_file = self.make_tar(mode='w:gz')
        tree = ControlDir.create_standalone_workingtree('tree')
        import_tar(tree, tar_file)
        self.assertTrue(tree.is_versioned('README'))

    def test_no_crash_with_bzrdir(self):
        tar_file = self.make_tar_with_bzrdir()
        tree = ControlDir.create_standalone_workingtree('tree')
        import_tar(tree, tar_file)
        # So long as it did not crash, that should be ok

    def test_get_archive_type(self):
        self.assertEqual(('tar', None), get_archive_type('foo.tar'))
        self.assertEqual(('zip', None), get_archive_type('foo.zip'))
        self.assertRaises(NotArchiveType, get_archive_type, 'foo.gif')
        self.assertEqual(('tar', 'gz'), get_archive_type('foo.tar.gz'))
        self.assertRaises(NotArchiveType, get_archive_type,
                          'foo.zip.gz')
        self.assertEqual(('tar', 'gz'), get_archive_type('foo.tgz'))
        self.assertEqual(('tar', 'lzma'), get_archive_type('foo.tar.lzma'))
        self.assertEqual(('tar', 'lzma'), get_archive_type('foo.tar.xz'))
        self.assertEqual(('tar', 'bz2'), get_archive_type('foo.tar.bz2'))


class TestWithStuff(TestCaseWithTransport):

    def transform_to_tar(self, tt):
        stream = BytesIO()
        export(tt.get_preview_tree(), root='', fileobj=stream, format='tar',
               dest=None)
        return stream

    def get_empty_tt(self):
        b = self.make_repository('foo')
        null_tree = b.revision_tree(_mod_revision.NULL_REVISION)
        tt = null_tree.preview_transform()
        tt.new_directory('', transform.ROOT_PARENT, b'tree-root')
        tt.fixup_new_roots()
        self.addCleanup(tt.finalize)
        return tt

    def test_nonascii_paths(self):
        self.requireFeature(UnicodeFilenameFeature)
        tt = self.get_empty_tt()
        tt.new_file('\u1234file', tt.root, [b'contents'], b'new-file')
        tt.new_file('other', tt.root, [b'contents'], b'other-file')
        tarfile = self.transform_to_tar(tt)
        tarfile.seek(0)
        tree = self.make_branch_and_tree('bar')
        import_tar(tree, tarfile)
        self.assertPathExists('bar/\u1234file')



# Copyright (C) 2020 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Management of hosted branches."""

from ... import version_info  # noqa: F401
from ...commands import plugin_cmds

plugin_cmds.register_lazy("cmd_gitlab_login", ["gl-login"], __name__ + ".cmds")


from ...forge import forges

forges.register_lazy("gitlab", __name__ + '.forge', "GitLab")


def test_suite():
    from unittest import TestSuite

    from .tests import test_suite
    result = TestSuite()
    result.addTest(test_suite())
    return result



# Copyright (C) 2005-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import datetime
import time

from breezy import errors
from breezy import revision as _mod_revision
from breezy.revisionspec import (InvalidRevisionSpec, RevisionInfo,
                                 RevisionSpec, RevisionSpec_dwim,
                                 RevisionSpec_tag)
from breezy.tests import TestCaseWithTransport


def spec_in_history(spec, branch):
    """A simple helper to change a revision spec into a branch search"""
    return RevisionSpec.from_string(spec).in_history(branch)


# Basic class, which just creates a really basic set of revisions
class TestRevisionSpec(TestCaseWithTransport):

    def setUp(self):
        super().setUp()
        # this sets up a revision graph:
        # r1: []             1
        # alt_r2: [r1]       1.1.1
        # r2: [r1, alt_r2]   2

        self.tree = self.make_branch_and_tree('tree')
        self.build_tree(['tree/a'])
        self.tree.lock_write()
        self.addCleanup(self.tree.unlock)
        self.tree.add(['a'])
        self.tree.commit('a', rev_id=b'r1')

        self.tree2 = self.tree.controldir.sprout('tree2').open_workingtree()
        self.tree2.commit('alt', rev_id=b'alt_r2')

        self.tree.merge_from_branch(self.tree2.branch)
        self.tree.commit('second', rev_id=b'r2')

    def get_in_history(self, revision_spec):
        return spec_in_history(revision_spec, self.tree.branch)

    def assertInHistoryIs(self, exp_revno, exp_revision_id, revision_spec):
        rev_info = self.get_in_history(revision_spec)
        self.assertEqual(exp_revno, rev_info.revno,
                         'Revision spec: %r returned wrong revno: %r != %r'
                         % (revision_spec, exp_revno, rev_info.revno))
        self.assertEqual(exp_revision_id, rev_info.rev_id,
                         'Revision spec: %r returned wrong revision id:'
                         ' %r != %r'
                         % (revision_spec, exp_revision_id, rev_info.rev_id))

    def assertInvalid(self, revision_spec, extra='',
                      invalid_as_revision_id=True):
        try:
            self.get_in_history(revision_spec)
        except InvalidRevisionSpec as e:
            self.assertEqual(revision_spec, e.spec)
            self.assertEqual(extra, e.extra)
        else:
            self.fail('Expected InvalidRevisionSpec to be raised for'
                      ' %r.in_history' % (revision_spec,))
        if invalid_as_revision_id:
            try:
                spec = RevisionSpec.from_string(revision_spec)
                spec.as_revision_id(self.tree.branch)
            except InvalidRevisionSpec as e:
                self.assertEqual(revision_spec, e.spec)
                self.assertEqual(extra, e.extra)
            else:
                self.fail('Expected InvalidRevisionSpec to be raised for'
                          ' %r.as_revision_id' % (revision_spec,))

    def assertAsRevisionId(self, revision_id, revision_spec):
        """Calling as_revision_id() should return the specified id."""
        spec = RevisionSpec.from_string(revision_spec)
        self.assertEqual(revision_id,
                         spec.as_revision_id(self.tree.branch))

    def get_as_tree(self, revision_spec, tree=None):
        if tree is None:
            tree = self.tree
        spec = RevisionSpec.from_string(revision_spec)
        return spec.as_tree(tree.branch)


class RevisionSpecMatchOnTrap(RevisionSpec):

    def _match_on(self, branch, revs):
        self.last_call = (branch, revs)
        return super()._match_on(branch, revs)


class TestRevisionSpecBase(TestRevisionSpec):

    def test_wants_no_revision_history(self):
        # If wants_revision_history = False, then _match_on should get None for
        # the branch revision history
        spec = RevisionSpecMatchOnTrap('foo', _internal=True)
        spec.in_history(self.tree.branch)

        self.assertEqual((self.tree.branch, None), spec.last_call)


class TestOddRevisionSpec(TestRevisionSpec):
    """Test things that aren't normally thought of as revision specs"""

    def test_none(self):
        self.assertInHistoryIs(None, None, None)

    def test_object(self):
        self.assertRaises(TypeError, RevisionSpec.from_string, object())


class RevisionSpec_bork(RevisionSpec):

    prefix = 'irrelevant:'

    def _match_on(self, branch, revs):
        if self.spec == "bork":
            return RevisionInfo.from_revision_id(branch, b"r1")
        else:
            raise InvalidRevisionSpec(self.spec, branch)


class TestRevisionSpec_dwim(TestRevisionSpec):

    # Don't need to test revno's explicitly since TRS_revno already
    # covers that well for us
    def test_dwim_spec_revno(self):
        self.assertInHistoryIs(2, b'r2', '2')
        self.assertAsRevisionId(b'alt_r2', '1.1.1')

    def test_dwim_spec_revid(self):
        self.assertInHistoryIs(2, b'r2', 'r2')

    def test_dwim_spec_tag(self):
        self.tree.branch.tags.set_tag('footag', b'r1')
        self.assertAsRevisionId(b'r1', 'footag')
        self.tree.branch.tags.delete_tag('footag')
        self.assertRaises(InvalidRevisionSpec, self.get_in_history, 'footag')

    def test_dwim_spec_tag_that_looks_like_revno(self):
        # Test that we slip past revno with things that look like revnos,
        # but aren't.  Tags are convenient for testing this since we can
        # make them look however we want.
        self.tree.branch.tags.set_tag('3', b'r2')
        self.assertAsRevisionId(b'r2', '3')
        self.build_tree(['tree/b'])
        self.tree.add(['b'])
        self.tree.commit('b', rev_id=b'r3')
        self.assertAsRevisionId(b'r3', '3')

    def test_dwim_spec_date(self):
        self.assertAsRevisionId(b'r1', 'today')

    def test_dwim_spec_branch(self):
        self.assertInHistoryIs(None, b'alt_r2', 'tree2')

    def test_dwim_spec_nonexistent(self):
        self.assertInvalid('somethingrandom', invalid_as_revision_id=False)
        self.assertInvalid('-1.1', invalid_as_revision_id=False)
        self.assertInvalid('.1', invalid_as_revision_id=False)
        self.assertInvalid('1..1', invalid_as_revision_id=False)
        self.assertInvalid('1.2..1', invalid_as_revision_id=False)
        self.assertInvalid('1.', invalid_as_revision_id=False)

    def test_append_dwim_revspec(self):
        original_dwim_revspecs = list(RevisionSpec_dwim._possible_revspecs)

        def reset_dwim_revspecs():
            RevisionSpec_dwim._possible_revspecs = original_dwim_revspecs
        self.addCleanup(reset_dwim_revspecs)
        RevisionSpec_dwim.append_possible_revspec(RevisionSpec_bork)
        self.assertAsRevisionId(b'r1', 'bork')

    def test_append_lazy_dwim_revspec(self):
        original_dwim_revspecs = list(RevisionSpec_dwim._possible_revspecs)

        def reset_dwim_revspecs():
            RevisionSpec_dwim._possible_revspecs = original_dwim_revspecs
        self.addCleanup(reset_dwim_revspecs)
        RevisionSpec_dwim.append_possible_lazy_revspec(
            "breezy.tests.test_revisionspec", "RevisionSpec_bork")
        self.assertAsRevisionId(b'r1', 'bork')


class TestRevisionSpec_revno(TestRevisionSpec):

    def test_positive_int(self):
        self.assertInHistoryIs(0, b'null:', '0')
        self.assertInHistoryIs(1, b'r1', '1')
        self.assertInHistoryIs(2, b'r2', '2')
        self.assertInvalid('3')

    def test_dotted_decimal(self):
        self.assertInHistoryIs(None, b'alt_r2', '1.1.1')
        self.assertInvalid('1.1.123')

    def test_negative_int(self):
        self.assertInHistoryIs(2, b'r2', '-1')
        self.assertInHistoryIs(1, b'r1', '-2')

        self.assertInHistoryIs(1, b'r1', '-3')
        self.assertInHistoryIs(1, b'r1', '-4')
        self.assertInHistoryIs(1, b'r1', '-100')

    def test_positive(self):
        self.assertInHistoryIs(0, b'null:', 'revno:0')
        self.assertInHistoryIs(1, b'r1', 'revno:1')
        self.assertInHistoryIs(2, b'r2', 'revno:2')

        self.assertInvalid('revno:3')

    def test_negative(self):
        self.assertInHistoryIs(2, b'r2', 'revno:-1')
        self.assertInHistoryIs(1, b'r1', 'revno:-2')

        self.assertInHistoryIs(1, b'r1', 'revno:-3')
        self.assertInHistoryIs(1, b'r1', 'revno:-4')

    def test_invalid_number(self):
        # Get the right exception text
        try:
            int('X')
        except ValueError as e:
            self.assertInvalid('revno:X', extra='\n' + str(e))
        else:
            self.fail()

    def test_missing_number_and_branch(self):
        self.assertInvalid('revno::',
                           extra='\ncannot have an empty revno and no branch')

    def test_invalid_number_with_branch(self):
        try:
            int('X')
        except ValueError as e:
            self.assertInvalid('revno:X:tree2', extra='\n' + str(e))
        else:
            self.fail()

    def test_non_exact_branch(self):
        # It seems better to require an exact path to the branch
        # Branch.open() rather than using Branch.open_containing()
        spec = RevisionSpec.from_string('revno:2:tree2/a')
        self.assertRaises(errors.NotBranchError,
                          spec.in_history, self.tree.branch)

    def test_with_branch(self):
        # Passing a URL overrides the supplied branch path
        revinfo = self.get_in_history('revno:2:tree2')
        self.assertNotEqual(self.tree.branch.base, revinfo.branch.base)
        self.assertEqual(self.tree2.branch.base, revinfo.branch.base)
        self.assertEqual(2, revinfo.revno)
        self.assertEqual(b'alt_r2', revinfo.rev_id)

    def test_int_with_branch(self):
        revinfo = self.get_in_history('2:tree2')
        self.assertNotEqual(self.tree.branch.base, revinfo.branch.base)
        self.assertEqual(self.tree2.branch.base, revinfo.branch.base)
        self.assertEqual(2, revinfo.revno)
        self.assertEqual(b'alt_r2', revinfo.rev_id)

    def test_with_url(self):
        url = self.get_url() + '/tree2'
        revinfo = self.get_in_history('revno:2:{}'.format(url))
        self.assertNotEqual(self.tree.branch.base, revinfo.branch.base)
        self.assertEqual(self.tree2.branch.base, revinfo.branch.base)
        self.assertEqual(2, revinfo.revno)
        self.assertEqual(b'alt_r2', revinfo.rev_id)

    def test_negative_with_url(self):
        url = self.get_url() + '/tree2'
        revinfo = self.get_in_history('revno:-1:{}'.format(url))
        self.assertNotEqual(self.tree.branch.base, revinfo.branch.base)
        self.assertEqual(self.tree2.branch.base, revinfo.branch.base)
        self.assertEqual(2, revinfo.revno)
        self.assertEqual(b'alt_r2', revinfo.rev_id)

    def test_different_history_lengths(self):
        # Make sure we use the revisions and offsets in the supplied branch
        # not the ones in the original branch.
        self.tree2.commit('three', rev_id=b'r3')
        self.assertInHistoryIs(3, b'r3', 'revno:3:tree2')
        self.assertInHistoryIs(3, b'r3', 'revno:-1:tree2')

    def test_invalid_branch(self):
        self.assertRaises(errors.NotBranchError,
                          self.get_in_history, 'revno:-1:tree3')

    def test_invalid_revno_in_branch(self):
        self.tree.commit('three', rev_id=b'r3')
        self.assertInvalid('revno:3:tree2')

    def test_revno_n_path(self):
        """Old revno:N:path tests"""
        wta = self.make_branch_and_tree('a')
        ba = wta.branch

        wta.commit('Commit one', rev_id=b'a@r-0-1')
        wta.commit('Commit two', rev_id=b'a@r-0-2')
        wta.commit('Commit three', rev_id=b'a@r-0-3')

        wtb = self.make_branch_and_tree('b')
        bb = wtb.branch

        wtb.commit('Commit one', rev_id=b'b@r-0-1')
        wtb.commit('Commit two', rev_id=b'b@r-0-2')
        wtb.commit('Commit three', rev_id=b'b@r-0-3')

        self.assertEqual((1, b'a@r-0-1'),
                         spec_in_history('revno:1:a/', ba))
        # The argument of in_history should be ignored since it is
        # redundant with the path in the spec.
        self.assertEqual((1, b'a@r-0-1'),
                         spec_in_history('revno:1:a/', None))
        self.assertEqual((1, b'a@r-0-1'),
                         spec_in_history('revno:1:a/', bb))
        self.assertEqual((2, b'b@r-0-2'),
                         spec_in_history('revno:2:b/', None))

    def test_as_revision_id(self):
        self.assertAsRevisionId(b'null:', '0')
        self.assertAsRevisionId(b'r1', '1')
        self.assertAsRevisionId(b'r2', '2')
        self.assertAsRevisionId(b'r1', '-2')
        self.assertAsRevisionId(b'r2', '-1')
        self.assertAsRevisionId(b'alt_r2', '1.1.1')

    def test_as_tree(self):
        tree = self.get_as_tree('0')
        self.assertEqual(_mod_revision.NULL_REVISION, tree.get_revision_id())
        tree = self.get_as_tree('1')
        self.assertEqual(b'r1', tree.get_revision_id())
        tree = self.get_as_tree('2')
        self.assertEqual(b'r2', tree.get_revision_id())
        tree = self.get_as_tree('-2')
        self.assertEqual(b'r1', tree.get_revision_id())
        tree = self.get_as_tree('-1')
        self.assertEqual(b'r2', tree.get_revision_id())
        tree = self.get_as_tree('1.1.1')
        self.assertEqual(b'alt_r2', tree.get_revision_id())


class TestRevisionSpec_revid(TestRevisionSpec):

    def test_in_history(self):
        # We should be able to access revisions that are directly
        # in the history.
        self.assertInHistoryIs(1, b'r1', 'revid:r1')
        self.assertInHistoryIs(2, b'r2', 'revid:r2')

    def test_missing(self):
        self.assertInvalid('revid:r3', invalid_as_revision_id=False)

    def test_merged(self):
        """We can reach revisions in the ancestry"""
        self.assertInHistoryIs(None, b'alt_r2', 'revid:alt_r2')

    def test_not_here(self):
        self.tree2.commit('alt third', rev_id=b'alt_r3')
        # It exists in tree2, but not in tree
        self.assertInvalid('revid:alt_r3', invalid_as_revision_id=False)

    def test_in_repository(self):
        """We can get any revision id in the repository"""
        # XXX: This may change in the future, but for now, it is true
        self.tree2.commit('alt third', rev_id=b'alt_r3')
        self.tree.branch.fetch(self.tree2.branch, b'alt_r3')
        self.assertInHistoryIs(None, b'alt_r3', 'revid:alt_r3')

    def test_unicode(self):
        """We correctly convert a unicode ui string to an encoded revid."""
        revision_id = '\N{SNOWMAN}'.encode()
        self.tree.commit('unicode', rev_id=revision_id)
        self.assertInHistoryIs(3, revision_id, 'revid:\N{SNOWMAN}')
        self.assertInHistoryIs(3, revision_id, 'revid:' +
                               revision_id.decode('utf-8'))

    def test_as_revision_id(self):
        self.assertAsRevisionId(b'r1', 'revid:r1')
        self.assertAsRevisionId(b'r2', 'revid:r2')
        self.assertAsRevisionId(b'alt_r2', 'revid:alt_r2')


class TestRevisionSpec_last(TestRevisionSpec):

    def test_positive(self):
        self.assertInHistoryIs(2, b'r2', 'last:1')
        self.assertInHistoryIs(1, b'r1', 'last:2')
        self.assertInHistoryIs(0, b'null:', 'last:3')

    def test_empty(self):
        self.assertInHistoryIs(2, b'r2', 'last:')

    def test_negative(self):
        self.assertInvalid('last:-1',
                           extra='\nyou must supply a positive value')

    def test_missing(self):
        self.assertInvalid('last:4')

    def test_no_history(self):
        tree = self.make_branch_and_tree('tree3')

        self.assertRaises(errors.NoCommits,
                          spec_in_history, 'last:', tree.branch)

    def test_not_a_number(self):
        last_e = None
        try:
            int('Y')
        except ValueError as e:
            last_e = e
        self.assertInvalid('last:Y', extra='\n' + str(last_e))

    def test_as_revision_id(self):
        self.assertAsRevisionId(b'r2', 'last:1')
        self.assertAsRevisionId(b'r1', 'last:2')


class TestRevisionSpec_before(TestRevisionSpec):

    def test_int(self):
        self.assertInHistoryIs(1, b'r1', 'before:2')
        self.assertInHistoryIs(1, b'r1', 'before:-1')

    def test_before_one(self):
        self.assertInHistoryIs(0, b'null:', 'before:1')

    def test_before_none(self):
        self.assertInvalid('before:0',
                           extra='\ncannot go before the null: revision')

    def test_revid(self):
        self.assertInHistoryIs(1, b'r1', 'before:revid:r2')

    def test_last(self):
        self.assertInHistoryIs(1, b'r1', 'before:last:1')

    def test_alt_revid(self):
        # This will grab the left-most ancestor for alternate histories
        self.assertInHistoryIs(1, b'r1', 'before:revid:alt_r2')

    def test_alt_no_parents(self):
        new_tree = self.make_branch_and_tree('new_tree')
        new_tree.commit('first', rev_id=b'new_r1')
        self.tree.branch.fetch(new_tree.branch, b'new_r1')
        self.assertInHistoryIs(0, b'null:', 'before:revid:new_r1')

    def test_as_revision_id(self):
        self.assertAsRevisionId(b'r1', 'before:revid:r2')
        self.assertAsRevisionId(b'r1', 'before:2')
        self.assertAsRevisionId(b'r1', 'before:1.1.1')
        self.assertAsRevisionId(b'r1', 'before:revid:alt_r2')


class TestRevisionSpec_tag(TestRevisionSpec):

    def make_branch_and_tree(self, relpath):
        # override format as the default one may not support tags
        return TestRevisionSpec.make_branch_and_tree(
            self, relpath, format='dirstate-tags')

    def test_from_string_tag(self):
        spec = RevisionSpec.from_string('tag:bzr-0.14')
        self.assertIsInstance(spec, RevisionSpec_tag)
        self.assertEqual(spec.spec, 'bzr-0.14')

    def test_lookup_tag(self):
        self.tree.branch.tags.set_tag('bzr-0.14', b'r1')
        self.assertInHistoryIs(1, b'r1', 'tag:bzr-0.14')
        self.tree.branch.tags.set_tag('null_rev', b'null:')
        self.assertInHistoryIs(0, b'null:', 'tag:null_rev')

    def test_failed_lookup(self):
        # tags that don't exist give a specific message: arguably we should
        # just give InvalidRevisionSpec but I think this is more helpful
        self.assertRaises(errors.NoSuchTag,
                          self.get_in_history,
                          'tag:some-random-tag')

    def test_as_revision_id(self):
        self.tree.branch.tags.set_tag('my-tag', b'r2')
        self.tree.branch.tags.set_tag('null_rev', b'null:')
        self.assertAsRevisionId(b'r2', 'tag:my-tag')
        self.assertAsRevisionId(b'null:', 'tag:null_rev')
        self.assertAsRevisionId(b'r1', 'before:tag:my-tag')


class TestRevisionSpec_date(TestRevisionSpec):

    def setUp(self):
        super(TestRevisionSpec, self).setUp()

        new_tree = self.make_branch_and_tree('new_tree')
        self.revid1 = new_tree.commit(
            'Commit one', timestamp=time.time() - 60 * 60 * 24)
        self.revid2 = new_tree.commit('Commit two')
        self.revid3 = new_tree.commit('Commit three')

        self.tree = new_tree

    def test_tomorrow(self):
        self.assertInvalid('date:tomorrow')

    def test_today(self):
        self.assertInHistoryIs(2, self.revid2, 'date:today')
        self.assertInHistoryIs(1, self.revid1, 'before:date:today')

    def test_yesterday(self):
        self.assertInHistoryIs(1, self.revid1, 'date:yesterday')

    def test_invalid(self):
        self.assertInvalid('date:foobar', extra='\ninvalid date')
        # You must have '-' between year/month/day
        self.assertInvalid('date:20040404', extra='\ninvalid date')
        # Need 2 digits for each date piece
        self.assertInvalid('date:2004-4-4', extra='\ninvalid date')

    def test_day(self):
        now = datetime.datetime.now()
        self.assertInHistoryIs(2, self.revid2,
                               'date:%04d-%02d-%02d' % (now.year, now.month, now.day))

    def test_as_revision_id(self):
        self.assertAsRevisionId(self.revid2, 'date:today')


class TestRevisionSpec_date_no_revno(TestRevisionSpec_date):

    # some formats don't implement .revno(), so it triggers a different codepath

    def get_in_history(self, revision_spec):
        old_revno = self.overrideAttr(self.tree.branch, 'revno', lambda: None)
        try:
            return spec_in_history(revision_spec, self.tree.branch)
        finally:
            self.tree.branch.revno = old_revno

    def test_today(self):
        self.assertInHistoryIs(2, self.revid2, 'date:today')
        # Drop before: since it messes with our monkeypatching of Branch.revno.


class TestRevisionSpec_ancestor(TestRevisionSpec):

    def test_non_exact_branch(self):
        # It seems better to require an exact path to the branch
        # Branch.open() rather than using Branch.open_containing()
        self.assertRaises(errors.NotBranchError,
                          self.get_in_history, 'ancestor:tree2/a')

    def test_simple(self):
        # Common ancestor of trees is 'alt_r2'
        self.assertInHistoryIs(None, b'alt_r2', 'ancestor:tree2')

        # Going the other way, we get a valid revno
        tmp = self.tree
        self.tree = self.tree2
        self.tree2 = tmp
        self.assertInHistoryIs(2, b'alt_r2', 'ancestor:tree')

    def test_self(self):
        self.assertInHistoryIs(2, b'r2', 'ancestor:tree')

    def test_unrelated(self):
        new_tree = self.make_branch_and_tree('new_tree')

        new_tree.commit('Commit one', rev_id=b'new_r1')
        new_tree.commit('Commit two', rev_id=b'new_r2')
        new_tree.commit('Commit three', rev_id=b'new_r3')

        # With no common ancestor, we should raise another user error
        self.assertRaises(errors.NoCommonAncestor,
                          self.get_in_history, 'ancestor:new_tree')

    def test_no_commits(self):
        new_tree = self.make_branch_and_tree('new_tree')
        self.assertRaises(errors.NoCommits,
                          spec_in_history, 'ancestor:new_tree',
                                           self.tree.branch)

        self.assertRaises(errors.NoCommits,
                          spec_in_history, 'ancestor:tree',
                                           new_tree.branch)

    def test_as_revision_id(self):
        self.assertAsRevisionId(b'alt_r2', 'ancestor:tree2')

    def test_default(self):
        # We don't have a parent to default to
        self.assertRaises(errors.NotBranchError, self.get_in_history,
                          'ancestor:')

        # Create a branch with a parent to default to
        tree3 = self.tree.controldir.sprout('tree3').open_workingtree()
        tree3.commit('foo', rev_id=b'r3')
        self.tree = tree3
        self.assertInHistoryIs(2, b'r2', 'ancestor:')


class TestRevisionSpec_branch(TestRevisionSpec):

    def test_non_exact_branch(self):
        # It seems better to require an exact path to the branch
        # Branch.open() rather than using Branch.open_containing()
        self.assertRaises(errors.NotBranchError,
                          self.get_in_history, 'branch:tree2/a')

    def test_simple(self):
        self.assertInHistoryIs(None, b'alt_r2', 'branch:tree2')

    def test_self(self):
        self.assertInHistoryIs(2, b'r2', 'branch:tree')

    def test_unrelated(self):
        new_tree = self.make_branch_and_tree('new_tree')

        new_tree.commit('Commit one', rev_id=b'new_r1')
        new_tree.commit('Commit two', rev_id=b'new_r2')
        new_tree.commit('Commit three', rev_id=b'new_r3')

        self.assertInHistoryIs(None, b'new_r3', 'branch:new_tree')

        # XXX: Right now, we use fetch() to make sure the remote revisions
        # have been pulled into the local branch. We may change that
        # behavior in the future.
        self.assertTrue(self.tree.branch.repository.has_revision(b'new_r3'))

    def test_no_commits(self):
        self.make_branch_and_tree('new_tree')
        self.assertRaises(errors.NoCommits,
                          self.get_in_history, 'branch:new_tree')
        self.assertRaises(errors.NoCommits,
                          self.get_as_tree, 'branch:new_tree')

    def test_as_revision_id(self):
        self.assertAsRevisionId(b'alt_r2', 'branch:tree2')

    def test_as_tree(self):
        tree = self.get_as_tree('branch:tree', self.tree2)
        self.assertEqual(b'r2', tree.get_revision_id())
        self.assertFalse(self.tree2.branch.repository.has_revision(b'r2'))


class TestRevisionSpec_submit(TestRevisionSpec):

    def test_submit_branch(self):
        # Common ancestor of trees is 'alt_r2'
        self.assertRaises(errors.NoSubmitBranch, self.get_in_history,
                          'submit:')
        self.tree.branch.set_parent('../tree2')
        self.assertInHistoryIs(None, b'alt_r2', 'submit:')
        self.tree.branch.set_parent('bogus')
        self.assertRaises(errors.NotBranchError, self.get_in_history,
                          'submit:')
        # submit branch overrides parent branch
        self.tree.branch.set_submit_branch('tree2')
        self.assertInHistoryIs(None, b'alt_r2', 'submit:')

    def test_as_revision_id(self):
        self.tree.branch.set_submit_branch('tree2')
        self.assertAsRevisionId(b'alt_r2', 'branch:tree2')


class TestRevisionSpec_mainline(TestRevisionSpec):

    def test_as_revision_id(self):
        self.assertAsRevisionId(b'r1', 'mainline:1')
        self.assertAsRevisionId(b'r2', 'mainline:1.1.1')
        self.assertAsRevisionId(b'r2', 'mainline:revid:alt_r2')
        spec = RevisionSpec.from_string('mainline:revid:alt_r22')
        e = self.assertRaises(
            InvalidRevisionSpec, spec.as_revision_id, self.tree.branch)
        self.assertContainsRe(str(e),
                              "Requested revision: 'mainline:revid:alt_r22' does not exist in"
                              " branch: ")

    def test_in_history(self):
        self.assertInHistoryIs(2, b'r2', 'mainline:revid:alt_r2')


class TestRevisionSpec_annotate(TestRevisionSpec):

    def setUp(self):
        super().setUp()
        self.tree = self.make_branch_and_tree('annotate-tree')
        self.build_tree_contents([('annotate-tree/file1', b'1\n')])
        self.tree.add('file1')
        self.tree.commit('r1', rev_id=b'r1')
        self.build_tree_contents([('annotate-tree/file1', b'2\n1\n')])
        self.tree.commit('r2', rev_id=b'r2')
        self.build_tree_contents([('annotate-tree/file1', b'2\n1\n3\n')])

    def test_as_revision_id_r1(self):
        self.assertAsRevisionId(b'r1', 'annotate:annotate-tree/file1:2')

    def test_as_revision_id_r2(self):
        self.assertAsRevisionId(b'r2', 'annotate:annotate-tree/file1:1')

    def test_as_revision_id_uncommitted(self):
        spec = RevisionSpec.from_string('annotate:annotate-tree/file1:3')
        e = self.assertRaises(
            InvalidRevisionSpec, spec.as_revision_id, self.tree.branch)
        self.assertContainsRe(str(e),
                              r"Requested revision: \'annotate:annotate-tree/file1:3\' does not"
                              " exist in branch: .*\nLine 3 has not been committed.")

    def test_non_existent_line(self):
        spec = RevisionSpec.from_string('annotate:annotate-tree/file1:4')
        e = self.assertRaises(InvalidRevisionSpec,
                              spec.as_revision_id, self.tree.branch)
        self.assertContainsRe(str(e),
                              r"Requested revision: \'annotate:annotate-tree/file1:4\' does not"
                              " exist in branch: .*\nNo such line: 4")

    def test_invalid_line(self):
        spec = RevisionSpec.from_string('annotate:annotate-tree/file1:q')
        e = self.assertRaises(InvalidRevisionSpec,
                              spec.as_revision_id, self.tree.branch)
        self.assertContainsRe(str(e),
                              r"Requested revision: \'annotate:annotate-tree/file1:q\' does not"
                              " exist in branch: .*\nNo such line: q")

    def test_no_such_file(self):
        spec = RevisionSpec.from_string('annotate:annotate-tree/file2:1')
        e = self.assertRaises(InvalidRevisionSpec,
                              spec.as_revision_id, self.tree.branch)
        self.assertContainsRe(str(e),
                              r"Requested revision: \'annotate:annotate-tree/file2:1\' does not"
                              " exist in branch: .*\nFile 'file2' is not versioned")

    def test_no_such_file_with_colon(self):
        spec = RevisionSpec.from_string('annotate:annotate-tree/fi:le2:1')
        e = self.assertRaises(InvalidRevisionSpec,
                              spec.as_revision_id, self.tree.branch)
        self.assertContainsRe(str(e),
                              r"Requested revision: \'annotate:annotate-tree/fi:le2:1\' does not"
                              " exist in branch: .*\nFile 'fi:le2' is not versioned")



# Copyright (C) 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os

from breezy import branch, builtins, errors, lock
from breezy.tests import TestCaseInTempDir, transport_util


class TestRevert(TestCaseInTempDir):

    def setUp(self):
        super().setUp()

    def test_revert_tree_write_lock_and_branch_read_lock(self):

        # install lock hooks to find out about cmd_revert's locking actions
        locks_acquired = []
        locks_released = []
        lock.Lock.hooks.install_named_hook('lock_acquired',
                                           locks_acquired.append, None)
        lock.Lock.hooks.install_named_hook('lock_released',
                                           locks_released.append, None)

        # execute the revert command (There is nothing to actually revert,
        # but locks are acquired either way.)
        revert = builtins.cmd_revert()
        revert.run()

        # make sure that only one lock is acquired and released.
        self.assertLength(1, locks_acquired)
        self.assertLength(1, locks_released)

        # make sure that the nonces are the same, since otherwise
        # this would not be the same lock.
        self.assertEqual(locks_acquired[0].details, locks_released[0].details)

        # make sure that the locks are checkout locks.
        self.assertEndsWith(locks_acquired[0].lock_url, "/checkout/lock")
        self.assertEndsWith(locks_released[0].lock_url, "/checkout/lock")



# Copyright (C) 2006 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Basic tests for AtomicFile"""

import os
import stat
import sys

from .. import atomicfile, osutils
from . import TestCaseInTempDir, TestSkipped


class TestAtomicFile(TestCaseInTempDir):

    def test_commit(self):
        f = atomicfile.AtomicFile('test')
        self.assertPathDoesNotExist('test')
        f.write(b'foo\n')
        f.commit()

        self.assertEqual(['test'], os.listdir('.'))
        self.check_file_contents('test', b'foo\n')
        self.assertRaises(atomicfile.AtomicFileAlreadyClosed, f.commit)
        self.assertRaises(atomicfile.AtomicFileAlreadyClosed, f.abort)
        # close is re-entrant safe
        f.close()

    def test_abort(self):
        f = atomicfile.AtomicFile('test')
        f.write(b'foo\n')
        f.abort()
        self.assertEqual([], os.listdir('.'))

        self.assertRaises(atomicfile.AtomicFileAlreadyClosed, f.abort)
        self.assertRaises(atomicfile.AtomicFileAlreadyClosed, f.commit)

        # close is re-entrant safe
        f.close()

    def test_close(self):
        f = atomicfile.AtomicFile('test')
        f.write(b'foo\n')
        # close on an open file is an abort
        f.close()
        self.assertEqual([], os.listdir('.'))

        self.assertRaises(atomicfile.AtomicFileAlreadyClosed, f.abort)
        self.assertRaises(atomicfile.AtomicFileAlreadyClosed, f.commit)

        # close is re-entrant safe
        f.close()

    def test_text_mode(self):
        f = atomicfile.AtomicFile('test', mode='wt')
        f.write(b'foo\n')
        f.commit()

        with open('test', 'rb') as f:
            contents = f.read()
        if sys.platform == 'win32':
            self.assertEqual(b'foo\r\n', contents)
        else:
            self.assertEqual(b'foo\n', contents)

    def can_sys_preserve_mode(self):
        # PLATFORM DEFICIENCY/ TestSkipped
        return sys.platform not in ('win32',)

    def _test_mode(self, mode):
        if not self.can_sys_preserve_mode():
            raise TestSkipped("This test cannot be run on your platform")
        f = atomicfile.AtomicFile('test', mode='wb', new_mode=mode)
        f.write(b'foo\n')
        f.commit()
        st = os.lstat('test')
        self.assertEqualMode(mode, stat.S_IMODE(st.st_mode))

    def test_mode_0666(self):
        self._test_mode(0o666)

    def test_mode_0664(self):
        self._test_mode(0o664)

    def test_mode_0660(self):
        self._test_mode(0o660)

    def test_mode_0640(self):
        self._test_mode(0o640)

    def test_mode_0600(self):
        self._test_mode(0o600)

    def test_mode_0400(self):
        self._test_mode(0o400)
        # Make it read-write again so cleanup doesn't complain
        os.chmod('test', 0o600)

    def test_no_mode(self):
        # The default file permissions should be based on umask
        umask = osutils.get_umask()
        f = atomicfile.AtomicFile('test', mode='wb')
        f.write(b'foo\n')
        f.commit()
        st = os.lstat('test')
        self.assertEqualMode(0o666 & ~umask, stat.S_IMODE(st.st_mode))

    def test_context_manager_commit(self):
        with atomicfile.AtomicFile('test') as f:
            self.assertPathDoesNotExist('test')
            f.write(b'foo\n')

        self.assertEqual(['test'], os.listdir('.'))
        self.check_file_contents('test', b'foo\n')

    def test_context_manager_abort(self):
        def abort():
            with atomicfile.AtomicFile('test') as f:
                f.write(b'foo\n')
                raise AssertionError
        self.assertRaises(AssertionError, abort)
        self.assertEqual([], os.listdir('.'))



# Copyright (C) 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#

"""Tests of the breezy.utextwrap."""

from .. import tests, utextwrap

# Japanese "Good morning".
# Each character have double width. So total 8 width on console.
_str_D = '\u304a\u306f\u3088\u3046'

_str_S = "hello"

# Combine single width characters and double width characters.
_str_SD = _str_S + _str_D
_str_DS = _str_D + _str_S


class TestUTextWrap(tests.TestCase):

    def check_width(self, text, expected_width):
        w = utextwrap.UTextWrapper()
        self.assertEqual(
            w._width(text),
            expected_width,
            "Width of %r should be %d" % (text, expected_width))

    def test_width(self):
        self.check_width(_str_D, 8)
        self.check_width(_str_SD, 13)

    def check_cut(self, text, width, pos):
        w = utextwrap.UTextWrapper()
        self.assertEqual((text[:pos], text[pos:]), w._cut(text, width))

    def test_cut(self):
        s = _str_SD
        self.check_cut(s, 0, 0)
        self.check_cut(s, 1, 1)
        self.check_cut(s, 5, 5)
        self.check_cut(s, 6, 5)
        self.check_cut(s, 7, 6)
        self.check_cut(s, 12, 8)
        self.check_cut(s, 13, 9)
        self.check_cut(s, 14, 9)
        self.check_cut('A' * 5, 3, 3)

    def test_split(self):
        w = utextwrap.UTextWrapper()
        self.assertEqual(list(_str_D), w._split(_str_D))
        self.assertEqual([_str_S] + list(_str_D), w._split(_str_SD))
        self.assertEqual(list(_str_D) + [_str_S], w._split(_str_DS))

    def test_wrap(self):
        self.assertEqual(list(_str_D), utextwrap.wrap(_str_D, 1))
        self.assertEqual(list(_str_D), utextwrap.wrap(_str_D, 2))
        self.assertEqual(list(_str_D), utextwrap.wrap(_str_D, 3))
        self.assertEqual(list(_str_D),
                         utextwrap.wrap(_str_D, 3, break_long_words=False))


class TestUTextFill(tests.TestCase):

    def test_fill_simple(self):
        # Test only can call fill() because it's just '\n'.join(wrap(text)).
        self.assertEqual("{}\n{}".format(_str_D[:2], _str_D[2:]),
                         utextwrap.fill(_str_D, 4))

    def test_fill_with_breaks(self):
        # Demonstrate complicated case.
        text = "spam ham egg spamhamegg" + _str_D + " spam" + _str_D * 2
        self.assertEqual('\n'.join(["spam ham",
                                     "egg spam",
                                     "hamegg" + _str_D[0],
                                     _str_D[1:],
                                     "spam" + _str_D[:2],
                                     _str_D[2:] + _str_D[:2],
                                     _str_D[2:]]),
                         utextwrap.fill(text, 8))

    def test_fill_without_breaks(self):
        text = "spam ham egg spamhamegg" + _str_D + " spam" + _str_D * 2
        self.assertEqual('\n'.join(["spam ham",
                                     "egg",
                                     "spamhamegg",
                                     # border between single width and double
                                     # width.
                                     _str_D,
                                     "spam" + _str_D[:2],
                                     _str_D[2:] + _str_D[:2],
                                     _str_D[2:]]),
                         utextwrap.fill(text, 8, break_long_words=False))

    def test_fill_indent_with_breaks(self):
        w = utextwrap.UTextWrapper(8, initial_indent=' ' * 4,
                                   subsequent_indent=' ' * 4)
        self.assertEqual('\n'.join(["    hell",
                                     "    o" + _str_D[0],
                                     "    " + _str_D[1:3],
                                     "    " + _str_D[3]
                                     ]),
                         w.fill(_str_SD))

    def test_fill_indent_without_breaks(self):
        w = utextwrap.UTextWrapper(8, initial_indent=' ' * 4,
                                   subsequent_indent=' ' * 4)
        w.break_long_words = False
        self.assertEqual('\n'.join(["    hello",
                                     "    " + _str_D[:2],
                                     "    " + _str_D[2:],
                                     ]),
                         w.fill(_str_SD))

    def test_fill_indent_without_breaks_with_fixed_width(self):
        w = utextwrap.UTextWrapper(8, initial_indent=' ' * 4,
                                   subsequent_indent=' ' * 4)
        w.break_long_words = False
        w.width = 3
        self.assertEqual('\n'.join(["    hello",
                                     "    " + _str_D[0],
                                     "    " + _str_D[1],
                                     "    " + _str_D[2],
                                     "    " + _str_D[3],
                                     ]),
                         w.fill(_str_SD))


class TestUTextWrapAmbiWidth(tests.TestCase):
    _cyrill_char = "\u0410"  # east_asian_width() == 'A'

    def test_ambiwidth1(self):
        w = utextwrap.UTextWrapper(4, ambiguous_width=1)
        s = self._cyrill_char * 8
        self.assertEqual([self._cyrill_char * 4] * 2, w.wrap(s))

    def test_ambiwidth2(self):
        w = utextwrap.UTextWrapper(4, ambiguous_width=2)
        s = self._cyrill_char * 8
        self.assertEqual([self._cyrill_char * 2] * 4, w.wrap(s))


# Regression test with Python's test_textwrap
# Note that some distribution including Ubuntu doesn't install
# Python's test suite.
try:
    from test import test_textwrap

    def override_textwrap_symbols(testcase):
        # Override the symbols imported by test_textwrap so it uses our own
        # replacements.
        testcase.overrideAttr(test_textwrap, 'TextWrapper',
                              utextwrap.UTextWrapper)
        testcase.overrideAttr(test_textwrap, 'wrap', utextwrap.wrap)
        testcase.overrideAttr(test_textwrap, 'fill', utextwrap.fill)

    def setup_both(testcase, base_class, reused_class):
        super(base_class, testcase).setUp()
        override_textwrap_symbols(testcase)
        reused_class.setUp(testcase)

    class TestWrap(tests.TestCase, test_textwrap.WrapTestCase):

        def setUp(self):
            setup_both(self, TestWrap, test_textwrap.WrapTestCase)

    class TestLongWord(tests.TestCase, test_textwrap.LongWordTestCase):

        def setUp(self):
            setup_both(self, TestLongWord, test_textwrap.LongWordTestCase)

    class TestIndent(tests.TestCase, test_textwrap.IndentTestCases):

        def setUp(self):
            setup_both(self, TestIndent, test_textwrap.IndentTestCases)


except ImportError:

    class TestWrap(tests.TestCase):  # type: ignore

        def test_wrap(self):
            raise tests.TestSkipped("test.test_textwrap is not available.")

    class TestLongWord(tests.TestCase):  # type: ignore

        def test_longword(self):
            raise tests.TestSkipped("test.test_textwrap is not available.")

    class TestIndent(tests.TestCase):  # type: ignore

        def test_indent(self):
            raise tests.TestSkipped("test.test_textwrap is not available.")



# Copyright (C) 2006 Canonical Ltd
# Authors:  Robert Collins <robert.collins@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for interface conformance of 'workingtree.get_parent_ids'"""

from breezy.tests import TestNotApplicable
from breezy.tests.per_workingtree import TestCaseWithWorkingTree


class TestGetParentIds(TestCaseWithWorkingTree):

    def test_get_parent_ids(self):
        t = self.make_branch_and_tree('t1')
        self.assertEqual([], t.get_parent_ids())
        rev1_id = t.commit('foo', allow_pointless=True)
        self.assertEqual([rev1_id], t.get_parent_ids())
        t2 = t.controldir.sprout('t2').open_workingtree()
        rev2_id = t2.commit('foo', allow_pointless=True)
        self.assertEqual([rev2_id], t2.get_parent_ids())
        t.merge_from_branch(t2.branch)
        self.assertEqual([rev1_id, rev2_id], t.get_parent_ids())
        for parent_id in t.get_parent_ids():
            self.assertIsInstance(parent_id, bytes)

    def test_pending_merges(self):
        """Test the correspondence between set pending merges and get_parent_ids."""
        wt = self.make_branch_and_tree('.')
        self.assertEqual([], wt.get_parent_ids())
        if not wt._format. supports_righthand_parent_id_as_ghost:
            raise TestNotApplicable(
                'format does not support right hand side parents '
                'that are ghosts')
        # the first pending merge replaces the 'last revision' because
        # 'last revision' is shorthand for 'left most parent'
        wt.add_pending_merge(b'foo@azkhazan-123123-abcabc')
        self.assertEqual([b'foo@azkhazan-123123-abcabc'], wt.get_parent_ids())
        # adding a merge which is already in the parents list gets ignored.
        wt.add_pending_merge(b'foo@azkhazan-123123-abcabc')
        self.assertEqual([b'foo@azkhazan-123123-abcabc'], wt.get_parent_ids())
        # adding a different merge results in it being appended to the list -
        # order is preserved.
        wt.add_pending_merge(b'wibble@fofof--20050401--1928390812')
        self.assertEqual([b'foo@azkhazan-123123-abcabc',
                          b'wibble@fofof--20050401--1928390812'],
                         wt.get_parent_ids())
